from .etc import mangle_name
from werkzeug.wrappers import Response
from flask.json import jsonify
from flask import request
import logging

LOG = logging.getLogger("idiotic.utils.api")

def jsonified(func):
    def decorator(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            LOG.exception("Exception encountered from API, args={}, kwargs={}".format(args, kwargs))
            return jsonify({"status": "error", "description": str(e)})
        return jsonify({"status": "success", "result": res})
    return decorator

def single_args(args):
    return {k: v[0] if isinstance(v, list) else v for k, v in args.items()}

class _APIWrapper:
    def __init__(self, api, module, base=None):
        self.__api = api
        self.module = module
        self.modname = mangle_name(getattr(module, "MODULE_NAME", module.__name__))
        if not base:
            base = join_url("/api/module", self.modname)
        self.path = base

    def serve(self, func, path, *args, get_args=False, get_form=False, get_data=False, content_type=None, raw_result=False, no_source=False, **kwargs):
        LOG.info("Adding API endpoint for {}: {} (content type {})".format(
            self.modname,
            path,
            content_type
        ))
        return self.__api.add_url_rule(path,
                                       "mod_{}_{}".format(
                                           self.modname,
                                           getattr(func, "__name__", "<unknown>")),
                                       _wrap_for_result(
                                           func, get_args, get_form,
                                           get_data, *args,
                                           content_type=content_type,
                                           raw_result=raw_result,
                                           no_source=no_source), **kwargs)

    def __getattr__(self, name):
        return getattr(self.__api, name)

def _wrap_for_result(func, get_args, get_form, get_data, no_source=False, content_type=None, raw_result=False, *args, **kwargs):
    def wrapper(*args, **kwargs):
        try:
            clean_get_args = {k: v[0] if isinstance(v, list) else v for k, v in getattr(request, "args", {}).items()}
            if get_args is True:
                kwargs.update(clean_get_args)
            elif get_args:
                kwargs[get_args] = clean_get_args

            clean_form = {k: v[0] if isinstance(v, list) else v for k, v in getattr(request, "form", {}).items()}
            if get_form is True:
                kwargs.update(clean_form)
            elif get_form:
                kwargs[get_form] = clean_form

            if get_data is True:
                kwargs["data"] = getattr(request, "data", "")
            elif get_data:
                kwargs[get_data] = getattr(request, "data", "")

            if not no_source:
                kwargs["source"] = "api"

            res = func(*args, **kwargs)
        except Exception as e:
            LOG.exception("Exception encountered from API, args={}, kwargs={}".format(args, kwargs))
            return jsonify({"status": "error", "description": str(e)})
        if content_type is None:
            if raw_result:
                return jsonify(res)
            else:
                return jsonify({"status": "success", "result": res})
        else:
            if isinstance(res, Response):
                return res
            else:
                return Response(res, mimetype=content_type)
    return wrapper

def join_url(*paths):
    return '/' + '/'.join((p.strip('/') for p in paths if p != '/'))



import datetime
import logging
import json
from .utils import IdioticEncoder

LOG = logging.getLogger("idiotic.event")

def pack_event(event):
    try:
        packed = {'__class__': type(event).__name__,
                  '__owner__': getattr(event, 'MODULE', 'unknown'),
                  '__kind__': 'event',
                  '_remote': True}
        packed.update(event.pack())
        return json.dumps(packed, cls=IdioticEncoder).encode('UTF-8')
    except AttributeError:
        try:
            packed = {
                '__class__': type(event).__name__,
                '__owner__': getattr(event, 'MODULE', 'unknown'),
                '_remote': True
            }
            packed.update(dict(event.__dict__))
            return json.dumps(packed, cls=IdioticEncoder).encode('UTF-8')
        except AttributeError:
            LOG.warn("Unable to pack event {} (type '{}') from module '{}'".format(
                str(event), type(event).__name__,
                getattr(event, 'MODULE', 'unknown')))
            return json.dumps({'__class__': type(event).__name__}).encode('UTF-8')

def unpack_event(data, modules):
    obj = json.loads(data.decode('UTF-8'))
    if '__owner__' in obj:
        owner = obj['__owner__']
        del obj['__owner__']

    if '__class__' in obj:
        clsname = obj['__class__']
        del obj['__class__']

    if owner == 'idiotic':
        cls = eval(clsname)
    else:
        cls = getattr(modules[owner], clsname)

    if cls is not None:
        return cls.unpack(obj)

class BaseEvent:
    @classmethod
    def unpack(cls, data):
        self = cls.__new__(cls)
        self.__dict__.update(data)
        return self

    def __init__(self):
        self.canceled = False
        self.time = datetime.datetime.now()

    def cancel(self):
        self.canceled = True

    def pack(self):
        res = {'__class__': type(self).__name__,
               '__owner__': getattr(self, 'MODULE', 'unknown'),
               '__kind__': 'event'}
        res.update(self.__dict__)
        return res

class SendStateChangeEvent(BaseEvent):
    MODULE = 'idiotic'
    def __init__(self, item, new, source):
        super().__init__()
        self.item = item
        self.new = new
        self.source = source

    def cancel(self):
        pass

class StateChangeEvent(BaseEvent):
    MODULE = 'idiotic'
    def __init__(self, item, old, new, source, kind):
        super().__init__()
        self.item = item
        self.old = old
        self.new = new
        self.source = source
        self.kind = kind

    def __repr__(self):
        return "StateChangeEvent({0.kind}, {0.old} -> {0.new} on {0.item} from {0.source})".format(self)

class SendCommandEvent(BaseEvent):
    MODULE = 'idiotic'
    def __init__(self, item, command, source="rule"):
        super().__init__()
        self.item = item
        self.command = command
        self.source = source

    def cancel(self):
        pass

class CommandEvent(BaseEvent):
    MODULE = 'idiotic'
    def __init__(self, item, command, source, kind):
        super().__init__()
        self.item = item
        self.command = command
        self.source = source
        self.kind = kind

    def __repr__(self):
        return "CommandEvent({0.kind}, '{0.command}' on {0.item} from {0.source})".format(self)

class SceneEvent(BaseEvent):
    MODULE = 'idiotic'
    def __init__(self, scene, state, kind):
        super().__init__()
        self.scene = scene
        self.state = state
        self.kind = kind

    def __repr__(self):
        return "SceneEvent({0.kind} {1} {0.scene}".format(self, "enter" if self.state else "leave")



import functools
import asyncio
import idiotic
import logging
import datetime
# FIXME this might not be needed, check if bind() is used anywhere
LOG = logging.getLogger("idiotic.rule")

def bind(func=None, *events):
    if len(events) == 0:
        events = [func]
        func = None

    if func:
        for event in events:
            event.bind(func)

            if not hasattr(func, "_rule_triggers"):
                func._rule_triggers = []
        func._rule_triggers.extend(list(events))
        return func
    else:
        def partial(func):
            return bind(func, *events)
        return partial

def augment(func=None, augmentation=None):
    if not augmentation:
        augmentation = func
        func = None

    if func:
        if not hasattr(func, "_rule_augments"):
            func._rule_augments = []

        func._rule_augments.append(augmentation)
        return augmentation.wrap(func)
    else:
        def partial(func):
            return augment(func, augmentation)
        return partial

class EventBinder:
    def bind(self, callback):
        raise NotImplementedError("You must override EventBinder.bind()")

    def get_filter(self):
        raise NotImplementedError("You must override EventBinder.get_filter()")

class Event(EventBinder):
    def __init__(self, type=None, *args, **kwargs):
        if type is not None:
            self.__filter = idiotic.utils.Filter(*args, type=type, **kwargs)
        else:
            self.__filter = idiotic.utils.Filter(*args, **kwargs)

    def bind(self, callback):
        from idiotic import instance
        instance.dispatcher.bind(callback, self.get_filter())

    def get_filter(self):
        return self.__filter

class Command(EventBinder):
    def __init__(self, item, command=None, time="after"):
        self.item = item
        if command:
            if isinstance(command, str):
                # Stupid strings and their also-being-iterable
                self.commands = [command]
            else:
                try:
                    self.commands = list(command)
                except TypeError:
                    self.commands = [command]
        else:
            self.commands = None

        if not time:
            time = "after"

        if time != "after" and time != "before" and time != "both":
            raise ValueError("argument 'time' to Command() must be either 'after', 'before', or 'both'")
        self.time = time

    def get_filter(self):
        kw = {}
        if self.commands is None and self.time == "both":
            pass
        elif not self.commands:
            kw["kind"] = self.time
        elif not self.time:
            kw["command__in"] = self.commands
        else:
            kw["command__in"] = self.commands
            kw["kind"] = self.time

        return idiotic.utils.Filter(type=idiotic.event.CommandEvent, **kw)

    def bind(self, callback):
        if self.commands is None and self.time == "both":
            self.item.bind_on_command(callback)
        elif not self.commands:
            self.item.bind_on_command(callback, kind=self.time)
        elif not self.time:
            self.item.bind_on_command(callback, command__in=self.commands)
        else:
            self.item.bind_on_command(callback, command__in=self.commands, kind=self.time)

class Change(EventBinder):
    def __init__(self, item, old=None, new=None, time="after"):
        self.item = item
        self.old = old
        self.new = new
        self.time = time

    def get_filter(self):
        kw = {}
        if self.old:
            kw['old'] = self.old
        if self.new:
            kw['new'] = self.new
        if self.time != "both":
            kw['kind'] = self.time
        return idiotic.utils.Filter(type=idiotic.event.StateChangeEvent, item=self.item, **kw)

    def bind(self, callback):
        kw = {}
        if self.old:
            kw['old'] = self.old
        if self.new:
            kw['new'] = self.new
        if self.time != "both":
            kw['kind'] = self.time
        self.item.bind_on_change(callback, **kw)

class Schedule(EventBinder):
    """A class for binding rules to cron-like schedules.

    This uses the schedule module to provide human-readable scheduling
    syntax.

    """
    def __init__(self, schedule):
        self.schedule = schedule

    def bind(self, callback):
        self.schedule.do(functools.partial(callback, None))
        #idiotic.scheduler.bind_to_schedule(schedule, callback)

    def get_filter(self):
        raise NotImplementedError("get_filter() is not supported on Schedule")

class EventAugmentation:
    def wrap(self, func):
        raise NotImplementedError("You must override EventAugmentation.wrap()")

class Delay(EventAugmentation):
    def __init__(self, binder, period=None, cancel=False, reset=True, consume=None):
        """Initialize a rule augmentation that delays the execution of a rule
when receives certain commands.

        Arguments:
        binder -- An EventBinder which determines when the rule will be
                  delayed. Any events which match this binder will be
                  delayed by period seconds before being sent to the rule.

        Keyword arguments:
        period -- How long, in seconds, to delay the event when binder is
                  matched.
        cancel -- When to cancel the pending event timer. If True, cancels
                  when any event other than binder is received. If False,
                  never cancels the pending event when it has started. If
                  an iterable of EventBinder, will cancel when any event
                  it contains is matched. If an EventBinder, will cancel
                  when an event matches it.
        reset --  When to reset the pending event timer, if it has not yet
                  completed running. By default, will reset when the event
                  received matches binder. If set to False, will never
                  reset the timer. If set to an iterable of EventBinder,
                  will reset when any event it contains is matched. If an
                  EventBinder, will reset when an event matches it.
        consume-- When to consume an event without passing it along to the
                  rule. This is distinct from cancel in that it will send
                  an event twice if it was delayed without being consumed.
                  This will only apply to events which were also matched
                  by binder or reset.

        Example:
        The most common use case for this augmentation is for preventing
        outputs from changing state rapidly when an input changes state.
        This is especially useful for something like a motion-sensor
        which controls a light. The following augmentation will delay
        responses to the "OFF" command of "motion_sensor" for 5 minutes,
        cancel the timer when the "ON" command is received, and reset the
        timer when the "OFF" command is received (if the timer is already
        running).
            Delay(Command(items.motion_sensor, "OFF"), period=300,
                          cancel=Command(items.motion_sensor, "ON"))
        """
        if binder != True and binder != False:
            self.filt = [binder.get_filter()] if isinstance(binder, EventBinder) else [b.get_filter() for b in binder]

        if cancel != True and cancel != False:
            # pylint: disable=no-member
            self.cancel = [cancel.get_filter()] if isinstance(cancel, EventBinder) else [c.get_filter() for c in cancel]
        else:
            self.cancel = cancel

        if reset != True and reset != False:
            # pylint: disable=no-member
            self.reset = [reset.get_filter()] if isinstance(reset, EventBinder) else [r.get_filter() for r in reset]
        else:
            self.reset = reset

        if consume is None:
            self.consume = self.filt
        elif consume != True and consume != False:
            self.consume = [consume.get_filter()] if isinstance(consume, EventBinder) else [c.get_filter() for c in consume]
        else:
            self.consume = consume

        self.period = period
        self.job = None

    def schedule(self, func, event):
        runtime = (datetime.datetime.now() + datetime.timedelta(seconds=self.period)).time()
        loop = asyncio.get_event_loop()
        self.job = loop.call_later(self.period, func, event)

    def cancel_job(self):
        if self.job:
            self.job.cancel()
            self.job = None

    def reschedule(self, func, event):
        self.cancel_job()
        self.schedule(func, event)

    def wrap(self, func):
        def wrapper(event, *args, **kwargs):
            # check if we should consume it
            consume = False
            if self.consume is True:
                consume = self.consume
            elif self.consume:
                for c in self.consume:
                    if c.check(event):
                        consume = True
                        break

            # Check whether event should be delayed
            for f in self.filt:
                if f.check(event):
                    # yes it should be delayed
                    if self.job:
                        # it was already scheduled
                        if self.reset is True:
                            # we always reset
                            self.reschedule(func, event)
                    else:
                        # it hasn't been scheduled yet
                        self.schedule(func, event)
                    if not consume:
                        return func(event, *args, **kwargs)
                    else:
                        return

            if self.job:
                # it's already scheduled so check if we should cancel
                if self.cancel is True:
                    # we always cancel
                    self.cancel_job()
                    if not consume:
                        return func(event, *args, **kwargs)

                elif self.cancel:
                    for c in self.cancel:
                        if c.check(event):
                            self.cancel_job()
                            if not consume:
                                return func(event, *args, **kwargs)

                if self.reset is True:
                    self.reschedule(func, event)
                    if not consume:
                        return func(event, *args, **kwargs)

                elif self.reset:
                    for r in self.reset:
                        if r.check(event):
                            self.reschedule(func, event)
                            if not consume:
                                return func(event, *args, **kwargs)

            # Don't forget to pass it through if we're not capturing it!
            if not consume:
                return func(event, *args, **kwargs)
        return wrapper

class DeDup(EventAugmentation):
    def __init__(self, binder=None, period=5, count=1, method="rolling"):
        """Combine multiple instances of the same event within a certain time
        period. NYI

        """
        LOG.warn("EventAugmentation DeDup: NOT YET IMPLEMENTED")

    def bind(self, func, *args, **kwargs):
        pass

    def wrap(self, func):
        return func



#!/usr/bin/env python3
"""idiotic Distributed Internet of Things Inhabitance Controller

Usage:
  idiotic.py --help
  idiotic.py --version
  idiotic.py [--base=<dir> | [--config=<file>] [--rules=<dir>] [--items=<dir>] [--modules=<dir>]] [-v | -vv | -q | -qq] [-s]

Options:
  -h --help           Show this text.
     --version        Print the version
  -v --verbose        Set verbosity.
  -q --quiet          Suppress output. Use -qq to suppress even errors.
  -b --base=<dir>     Path to idiotic config base directory [default: /etc/idiotic].
  -c --config=<file>  Path to idiotic config file [default: <base>/conf.json].
  -r --rules=<dir>    Path to rules config directory [default: <base>/rules].
  -i --items=<dir>    Path to items config directory [default: <base>/items].
  -m --modules=<dir>  Path to modules directory [default: <base>/modules].
  -s --standalone     Run without connecting to other instances.
"""

import os
import sys
import json
import time
import docopt
import socket
import asyncio
import logging
import threading
import aiohttp.wsgi
import idiotic
from idiotic import utils, item, rule, distrib, event
# FIXME: This is sort of a hack, due to dependency resolution order
# problems (persistence and distrib must import idiotic for the
# registration hooks). The better solution would be to move concrete
# persistence and distribution methods into a lib folder
import idiotic.persistence
#import idiotic.distrib

LOG = logging.getLogger("idiotic.main")

def init():
    # load command-line options
    arguments = docopt.docopt(__doc__, version="Current version")

    # All these dashes are stupid
    arguments = {k.lstrip('--'): v for k,v in arguments.items()}
    # A little strange, but this should work correctly for everything.
    arguments = {k: os.path.join(arguments["base"],v.replace("<base>/","")) if type(v) is str and "<base>" in v else v for k,v in arguments.items()}

    verbose = arguments.get("verbose", 0)
    quiet = arguments.get("quiet", 0)
    level = logging.INFO

    if verbose == 1:
        level = logging.DEBUG
    elif verbose == 2:
        level = logging.DEBUG
    elif quiet == 1:
        level = logging.ERROR
    elif quiet == 2:
        level = logging.CRITICAL + 1

    logging.basicConfig(level=level)

    config = {}
    name = socket.gethostname()

    try:
        with open(arguments["config"]) as conf_file:
            config = json.load(conf_file)

            if "name" in config:
                name = config["name"]

            if "modules" in config:
                config["modules"].update({"builtin": {"api_base": "/"}})
            else:
                config["modules"] = {"builtin": {"api_base": "/"}}

    except (OSError, IOError):
        LOG.exception("Could not load config file {}:".format(arguments["config"]))

    instance = idiotic.Idiotic(config, name)

    # FIXME global state
    idiotic.instance = instance

    # load modules
    LOG.info("Loading modules from {}".format(arguments["modules"]))
    for module, assets in utils.load_dir(arguments["modules"], True):
        instance.augment_module(module)
        if module.__name__.startswith("_"):
            instance._register_builtin_module(module, assets)
        else:
            instance._register_module(module, assets)

    # load items
    LOG.info("Loading items from {}".format(arguments["items"]))
    for module, _ in utils.load_dir(arguments["items"]):
        instance.augment_module(module)

    # load rules
    LOG.info("Loading rules from {}".format(arguments["rules"]))
    for module, _ in utils.load_dir(arguments["rules"]):
        instance.augment_module(module)

    for module in instance.modules.all(lambda m:hasattr(m, "ready")):
        module.ready()

    # correspond with other instances?
    if "distribution" in config and config["distribution"]:
        LOG.info("Initializing distribution system...")
        if "method" in config["distribution"]:
            # Built-in methods go here
            instance._start_distrib(config["distribution"]["method"],
                                   name, config["distribution"])
        else:
            LOG.warn("No distribution method specified. Skipping.")
    else:
        LOG.debug("Not setting up distribution.")

    # connect to persistence engine
    if "persistence" in config and config["persistence"]:
        LOG.info("Connecting to persistence engine...")
        if not config["persistence"].get("disabled", False):
            if "method" in config["persistence"]:
                try:
                    instance._start_persistence(config["persistence"]["method"], config["persistence"])
                except NameError as e:
                    LOG.error("Could not locate persistence engine '{}' -- check spelling?".format(config["persistence"]["method"]), e)
            else:
                LOG.warn("Persistence engine not specified. Skipping.")
        else:
            LOG.warn("Persistence is disabled.")
    else:
        LOG.info("Not setting up persistence.")

    return instance

def shutdown(instance):
    instance._stop_distrib()

    instance._stop_persistence()

    logging.shutdown()

@asyncio.coroutine
def run_everything(*things):
    while True:
        for c in things:
            yield from c()
        yield from asyncio.sleep(.01)

def main():
    instance = init()

    try:
        loop = asyncio.get_event_loop()

        api_config = instance.config.get("api", {})
        listen = api_config.get("listen", "*")
        port = api_config.get("port", 5000)

        # FIXME find out why this is necessary
        instance._finalize_api()

        server = loop.create_server(lambda: aiohttp.wsgi.WSGIServerHttpProtocol(instance.api, readpayload=True), listen, port)
        loop.run_until_complete(asyncio.gather(instance.run_scheduled_jobs(),
                                               server,
                                               instance.dispatcher.run()))
    except KeyboardInterrupt:
        LOG.info("Shutting down")
        shutdown(instance)

if __name__ == '__main__':
    main()



import json

# Expects units to be a dictionary like:
# { "a": {"depends": ["b", "c"], "value": ...},
#   "b": {"depends": ["c"], "value": ...},
#   "c": {"depends": [], "value": ...},
#   "d": {"depends": ["c"], "value": ...} }
def resolve_depends(units, preresolved=[]):
    for name in units:
        units[name]["depends"] = set(units[name].get("depends", []))
        units[name]["depends"] |= set([u for u in units[name].get("optdepends", []) if u in units])

    # Contains dependencies resolved last round
    resolved = set(preresolved)
    while units:
        # Mark as resolved, to handle next time around
        to_resolve = set()
        for k, v in list(units.items()):
            # Remove dependencies we resolved already
            v["depends"] -= resolved

            # If this has dependencies still, skip it
            if v["depends"]: continue

            # Otherwise, mark this as resolved
            to_resolve.add(k)

            # And yield it to the caller
            yield (k, units.pop(k).get("value", None))

        resolved = to_resolve

        if not resolved:
            # We have not resolved anything this round
            # This means we're stuck

            present = set(units.keys()) | set(preresolved)

            required = set()
            for v in units.values():
                required.update(v["depends"])

            if required - present:
                # There is an unmet dependency
                raise ValueError("Unsatisfied dependencies: " + ", ".join(list(required)))
            else:
                # There is a circular dependency
                raise ValueError("Circular dependency detected")

def load_metadata(path):
    try:
        with open(path) as f:
            data = json.load(f)
    except:
        data = {}

    if "name" not in data:
        data["name"] = os.path.splitext(os.path.basename(path))[0]

    return data

def preload_modules(path, system=False):
    modules = {}
    for path, assets in load_dir(path, True, False):
        if path.endswith(".meta"):
            metadata = load_metadata(path)
            if "name" not in metadata:
                pass

def load_dir(path, include_assets=False, do_import=True):
    sys.path.insert(0, os.path.abspath("."))
    modules = []
    for f in os.listdir(path):
        try:
            if f.startswith(".") or f.endswith("~") or f.endswith("#") or f.startswith("__"):
                continue

            LOG.info("Loading file {}...".format(os.path.join(path, f)))
            name = os.path.splitext(f)[0]

            try:
                if do_import:
                    modules.append((imp.load_source(name, os.path.join(path, f)), None))
                else:
                    modules.append((os.path.join(path, f), None))

            except IsADirectoryError:
                LOG.info("Attempting to load directory {} as a module...".format(
                    os.path.join(path, f)))

                try:
                    if do_import:
                        mod = imp.load_source(name, os.path.join(path, f, '__init__.py'))
                    else:
                        mod = os.path.join(path, f, '__init__.py')
                    assets = None
                    if os.path.exists(os.path.join(path, f, 'assets')) and \
                       os.path.isdir(os.path.join(path, f, 'assets')):
                        assets = os.path.abspath(os.path.join(path, f, 'assets'))

                    modules.append((mod, assets))
                except FileNotFoundError:
                    LOG.error("Unable to load module {}: {} does not exist".format(
                        name, os.path.join(path, f, '__init__.py')))
        except:
            LOG.exception("Exception encountered while loading {}".format(os.path.join(path, f)))

    return modules



import idiotic
import logging

LOG = logging.getLogger("idiotic.distrib.base")

class RemoteItem:
    def __init__(self, neighbor, name):
        pass

class RemoteModule:
    def __init__(self, neighbor, name):
        pass

class Neighbor:
    def __init__(self, config):
        pass

class DistributionType(type):
    def __new__(mcs, name, bases, attrs):
        if name.startswith('None'):
            return None

        newattrs = dict(attrs)
        if 'NAME' not in attrs:
            newattrs['NAME'] = name

        return super(DistributionType, mcs).__new__(mcs, name, bases, newattrs)

    def __init__(cls, name, bases, attrs):
        super(DistributionType, cls).__init__(name, bases, attrs)
        if name != "TransportMethod":
            idiotic._register_distrib_type(cls.NAME, cls)

class TransportMethod(metaclass=DistributionType):
    NEIGHBOR_CLASS = RemoteItem
    MODULE_CLASS = RemoteModule
    ITEM_CLASS = Neighbor

    def __init__(self, hostname, config):
        raise NotImplementedError("Cannot use abstract transport")

    def send(self, event, targets=True):
        """Send packed data to all or a subset of this node's neighbors.

        """

    def receive(self, cb, cancel=False):
        """Add a callback that will be called for all packed data received. If
'cancel' is True, will instead cancel the passed callback.

        """
        if not hasattr(self, "callbacks"):
            self.callbacks = set()

        if cancel:
            self.callbacks.remove(cb)
        else:
            self.callbacks.add(cb)

    def run(self):
        """Begin running any necessary loop for running the transport
method.

        """

    def stop(self):
        """Stops running any loops in preparation for shutdown."""

    def connect(self):

        """Connect to the main server, if applicable, and all configured or
discovered neighbors as needed.

        """

    def disconnect(self):
        """Disconnect from the main server, if applicable, and all configured
or discovered neighbors as needed.

        """

    def reconnect(self):
        self.disconnect()
        self.connect()

    def neighbors(self):
        """Return a list of neighbors which are currently connected with this
        node.

        """



import functools
import datetime
import logging
import idiotic
from idiotic import event, history, utils

LOG = logging.getLogger("idiotic.item")

def command(func):
    def command_decorator(self, *args, **kwargs):
        # If we get passed a source (e.g., UI, Rule, Binding), consume
        # it so we don't break our child function
        if "source" in kwargs:
            source = kwargs["source"]
            del kwargs["source"]
        else:
            source = None

        if "command" in kwargs:
            name = kwargs["command"]
        else:
            name = func.__name__

        LOG.debug("@command({}) on {}".format(name, self))

        if not self.enabled:
            LOG.info("Ignoring command {} on disabled item {}".format(name, self))
            return

        # Create an event and send it
        pre_event = event.CommandEvent(self, name, source, kind="before")
        self.idiotic.dispatcher.dispatch(pre_event)

        if not pre_event.canceled:
            func(self, *args, **kwargs)

            if hasattr(self, "__command_history"):
                self.__command_history.record(name)

            if self.idiotic.persist_instance:
                self.idiotic.persist_instance.append_item_history(
                    self, datetime.datetime.now(),
                    name, kind="command",
                    extra={"args": args, "kwargs": kwargs} if args or kwargs else None)

            post_event = event.CommandEvent(self, name, source, kind="after")
            self.idiotic.dispatcher.dispatch(post_event)
    return command_decorator

class BaseItem:
    """The base class for an item which implements all the basic
    behind-the-scenes functionality and makes no assumptions about the
    nature of its state.

    """
    def __init__(self, name, groups=None, friends=None, bindings=None, update=None, tags=None, ignore_redundant=False, aliases=None, id=None):
        self.name = name
        self._state = None

        if id is None:
            self.id = utils.mangle_name(name)
        else:
            self.id = utils.mangle_name(id)

        self.ignore_redundant = ignore_redundant

        if tags is None:
            self.tags = set()
        else:
            self.tags = set(tags)

        if friends is None:
            self.friends = {}
        else:
            self.friends = friends

        if groups is None:
            self.groups = set()
        else:
            self.groups = set(groups)

        for group in self.groups:
            group.add(self)

        if aliases is None:
            self.aliases = {}
        else:
            self.aliases = aliases

        self.enabled = True

        self.__command_history = history.History()
        self.__state_history = history.History()

        self.__state_overlay = []

        idiotic.instance._register_item(self)

        if bindings:
            for module_name, args in bindings.items():
                LOG.debug("Setting {} bindings on {}".format(module_name, self))
                try:
                    module = self.idiotic.modules[module_name]
                except NameError:
                    LOG.warning("Module '{}' not found -- skipping".format(module_name))
                else:
                    module.bind_item(self, **args)

        if update:
            def wrap_update(item, attr, base_func):
                if attr:
                    setattr(item, attr, base_func(item))
                else:
                    base_func(item)

            if isinstance(update, dict):
                for key, updaters in update.items():
                    for interval, func in updaters:
                        interval.do(wrap_update, self, key, func)
            elif isinstance(update, tuple):
                update[0].do(wrap_update, self, None, update[1])

    def bind_on_command(self, function, **kwargs):
        LOG.debug("Binding on command for {}".format(self))
        self.idiotic.dispatcher.bind(function, utils.Filter(type=event.CommandEvent, item=self, **kwargs))

    def bind_on_change(self, function, **kwargs):
        self.idiotic.dispatcher.bind(function, utils.Filter(type=event.StateChangeEvent, item=self, **kwargs))

    def __str__(self):
        return type(self).__name__ + " '" + self.name + "'"

    def __repr__(self):
        return type(self).__name__ + " '" + self.name + "' on local"

    def disable(self):
        self.enabled = False

    def enable(self):
        self.enabled = True

    def has_tag(self, tag):
        return tag and tag.lower() in self.tags

    def add_tag(self, tag):
        self.tags.add(tag)

    def remove_tag(self, tag):
        self.tags.remove(tag)

    def change_state(self, state):
        pass

    def overlay_state(self, state, tag=None, disable=False):
        self.__state_overlay.append({"state": state, "disabled": disable, "tag": tag})
        self.__compute_state_overlay()

    def remove_state_overlay(self, tag=None):
        if tag:
            for i, overlay in enumerate(self.__state_overlay):
                if overlay["tag"] == tag:
                    self.__state_overlay = self.__state_overlay[:i] + self.__state_overlay[i+1:]
                    break
        else:
            self.__state_overlay.pop()
        self.__compute_state_overlay()

    def __compute_state_overlay(self):
        target_state = None
        enabled = None
        for overlay in reversed(self.__state_overlay):
            if "state" in overlay and target_state is None:
                target_state = overlay["state"]

            if "disabled" in overlay and enabled is None:
                enabled = not overlay["disabled"]

        if target_state is not None:
            self.change_state(target_state)
        elif not self.__state_overlay:
            self.change_state(self.state)

        if enabled is not None:
            self.enabled = enabled
        else:
            self.enabled = True

    @property
    def state(self):
        if self.__state_overlay:
            return self.__state_overlay[-1]["state"]
        return self._state

    @state.setter
    def state(self, state):
        self._set_state_from_context(state)

    def command(self, name, *args, **kwargs):
        if name in self.aliases:
            name = self.aliases[name]

        if hasattr(self, name) and callable(getattr(self, name)):
            return getattr(self, name)(*args, **kwargs)
        else:
            raise ValueError("Command {} on item {} does not exist or is not a command".format(name, self))

    def _set_state_from_context(self, val, source="rule"):
        if self.__state_overlay:
            return

        if not self.enabled:
            LOG.info("Ignoring state change on disabled item {}".format(self))
            return

        # We don't send an event if there has been literally no change
        if self._state == val and self.ignore_redundant:
            LOG.debug("Ignoring redundant state change for {}".format(self))
            return

        if self._state != val:
            LOG.info("{} changed state from {} -> {}".format(self, self._state, val))

        old = self._state
        pre_event = event.StateChangeEvent(self, old, val, source, kind="before")
        self.idiotic.dispatcher.dispatch(pre_event)
        if not pre_event.canceled:
            self._state = val

            if hasattr(self, "__state_history"):
                self.__state_history.record(self._state)

            for group in self.groups:
                group._member_state_changed(self, self._state, source)

            post_event = event.StateChangeEvent(self, old, val, source, kind="after")
            self.idiotic.dispatcher.dispatch(post_event)

    @property
    def state_history(self):
        return self.__state_history

    @property
    def command_history(self):
        return self.__command_history

    def commands(self):
        return [k for k in dir(self) if callable(getattr(self, k, None)) and getattr(self, k).__name__ == "command_decorator"]

    def pack(self):
        res = {
            "__class__": type(self).__name__,
            "__owner__": getattr(self, 'MODULE', 'unknown'),
            "__kind__": "item",
            "__host__": None,
            "__commands__": self.commands(),
            "__attrs__": [k for k in dir(self) if not callable(getattr(self, k, None))
                          and not k.startswith('_')],
            "__methods__": [k for k in dir(self) if callable(getattr(self, k, None))
                            and not k.startswith('_')]
        }

        return res

    def json(self):
        res = {
            "type": type(self).__name__,
            "name": self.name,
            "id": getattr(self, "id", None),
            "tags": list(self.tags),
            "commands": self.commands(),
            "methods": [k for k in dir(self) if callable(getattr(self, k, None))
                        and not k.startswith('_')],
            "aliases": self.aliases,
        }

        if hasattr(self, "state"):
            res["state"] = self.state

        return res


class ItemProxy(BaseItem):
    def __init__(self, idiotic, typename, host, name, commands, attrs, methods,
                 ignore_redundant=False):
        self.typename = typename
        self.host = host
        self.name = name
        self.commands = commands
        self.attrs = attrs
        self.methods = methods
        self._state = None

        self.ignore_redundant = ignore_redundant

        self.idiotic.dispatcher.bind(self.__cache_update, idiotic.utils.Filter(
            item=self.name, type=event.StateChangeEvent))

    def pack(self):
        res = {
            "__class__": self.typename,
            "__host__": self.host,
            "__commands__": self.commands,
            "__attrs__": self.attrs,
            "__methods__": self.methods,
        }

        res.update(self.__dict__)

        return res

    def __cache_update(self, e):
        self._state = e.new

    def _set_state_from_context(self, val, source="rule"):
        if self._state == val and self.ignore_redundant:
            LOG.debug("Ignoring redundant state change for {}".format(self))
            return

        LOG.info("signaling change state on {} from {} -> {}".format(
            self, self._state, val))

        self.idiotic.dispatcher.dispatch(event.SendStateChangeEvent(self.name, val, source))

    def __getattr__(self, attr):
        if attr in self.commands:
            return functools.partial(self.idiotic.dispatcher.dispatch,
                                     event.SendCommandEvent,
                                     source = None)
        elif attr in self.attrs:
            if attr in self._cache:
                return self._cache[attr]
            else:
                raise NotImplementedError("Remote items do not yet support attribute access")

    def __setattr__(self, attr, val):
        if attr in self.attrs:
            self.idiotic.dispatcher.dispatch(event.SendStateChangeEvent(self.name, val, None))
        else:
            raise NameError("Item has no attribute {}".format(attr))

    def __repr__(self):
        return "proxy for " + self.typename + " '" + self.name + "' on " + self.host

    def __eq__(self, other):
        return (isinstance(other, BaseItem) and self.name == other.name) or \
            isinstance(other, str) and other == self.name

    def __req__(self, lhs):
        return self.__eq__(lhs)

class Toggle(BaseItem):
    """An item which has two discrete states between which it may be
    toggled, and which is not affected by repeated identical commands.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def change_state(self, state):
        if state:
            self.on()
        else:
            self.off()
        # otherwise it's already ok

    def _set_state_from_context(self, val, source="rule"):
        if val in ("True", "true", "1", "on"):
            val = True
        elif val in ("False", "false", "0", "off"):
            val = False
        super()._set_state_from_context(bool(val), source)

    @command
    def on(self):
        self.state = True

    @command
    def off(self):
        self.state = False

    def toggle(self, *args, **kwargs):
        if self.state:
            self.off(*args, **kwargs)
        else:
            self.on(*args, **kwargs)

class Trigger(BaseItem):
    """An item with no state, but which may be activated repeatedly,
    triggering a distinct command each time.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @command
    def trigger(self):
        pass

class Number(BaseItem):
    """An item which represents a numerical quantity of some sort."""

    def __init__(self, *args, kind=float, **kwargs):
        self.kind = kind
        super().__init__(*args, **kwargs)

    def change_state(self, state):
        self.set(state)

    @command
    def set(self, val):
        try:
            self.state = self.kind(val)
        except (ValueError, TypeError):
            LOG.warn("Invalid {} argument to Number.set: {}".format(self.kind.__name__, val))

class Text(BaseItem):
    """An item which represents a blob of text."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def change_state(self, state):
        self.set(str(state))

    @command
    def set(self, val):
        self.state = str(val)

class Motor(BaseItem):
    """An item which can move forward, reverse, and stop."""

    # Options for the current state.
    MOVING_FORWARD = "MOVING_FORWARD"
    MOVING_REVERSE = "MOVING_REVERSE"
    STOPPED = "STOPPED"
    STOPPED_START = "STOPPED_START"
    STOPPED_END = "STOPPED_END"

    STATES = (MOVING_FORWARD,
              MOVING_REVERSE,
              STOPPED)

    STATES_CONSTRAINED = STATES + (STOPPED_START, STOPPED_END)

    def __init__(self, *args, constrained=False, timeout=None, **kwargs):
        self.constrained = constrained
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def change_state(self, state):
        if state == Motor.MOVING_FORWARD or state == Motor.STOPPED_END:
            self.forward()
        elif state == Motor.MOVING_REVERSE or state == Motor.STOPPED_START:
            self.reverse()
        elif state == Motor.STOPPED:
            self.stop()

    @command
    def forward(self):
        if self.state != Motor.STOPPED_END or not self.constrained:
            self.state = Motor.MOVING_FORWARD
            if self.timeout:
                raise NotImplementedError("timeout is not implemented. probably should do it with asyncio, or implement timers")
        else:
            LOG.debug("Not moving {} forward; already at end stop".format(self))

    @command
    def reverse(self):
        if self.state != Motor.STOPPED_START or not self.constrained:
            self.state = Motor.MOVING_REVERSE
            if self.timeout:
                raise NotImplementedError("timeout is not implemented. probably should do it with asyncio, or implement timers")
        else:
            LOG.debug("Not moving {} reverse; already at start stop".format(self))

    @command
    def stop(self):
        self.state = Motor.STOPPED

class _BagOfHolding:
    def __contains__(self, arg):
        return True

class _Sieve:
    def __contains__(self, arg):
        return False

class _ImposterDict:
    def __init__(self, item):
        self.item = item

    def __contains__(self, arg):
        return True

    def __getitem__(self, index):
        return self.item

class Group(BaseItem):
    """An item which contains other items. It may have custom behavior
    defined to facilitate acting on all its members at once, and to
    summarize its state.

    """
    def __init__(self, *args, state=any, state_set=None, commands=False, command_send=False, members=None, **kwargs):
        """Initialize a Group item, which may or may not handle state updates
        and commands in a custom manner.

        Keyword Arguents:
        state      -- A function which will be used to compute the
                      group's state. It should accept an iterable of
                      BaseItems and return the state. The default is
                      the builtin `any`. If set to `None`, the group
                      will have its own state, indepedent of that
                      of its members.
        state_set  -- A function which will be called when the state
                      of the group is set. It should accept an
                      iterable of BaseItems, and the new state value.
                      This is generally used to set the state of a
                      group's members by changing only the group's
                      state. By default, will do nothing. This
                      argument will be ignored if the 'state'
                      argument is `None`.
        commands   -- How to handle commands sent to the group's
                      members. If `True`, all commands on members of
                      the group will also be sent to the group. If
                      `False`, the group will only receive its own
                      commands. If this argument is an iterable, only
                      commands named by its elements will be passed on
                      to the group, and all others will be ignored.
                      Defaults to `False`.
        command_send -How to handle commands sent to the group itself.
                      If False, commands sent to the group will have
                      no effect. If this is a callable, it will be
                      passed an iterable of BaseItems, the name of the
                      command, and any arguments the command may
                      take. If this is a dictionary, the keys should
                      correspond to command names, while the values
                      are callables for which the commands will be
                      called. The function will be called in the same
                      manner as described previously. Otherwise, this
                      may be set to `True` to automatically call the
                      command on all members of the group.

        """
        super().__init__(*args, **kwargs)

        self.members = members or []

        for item in self.members:
            if self not in item.groups:
                item.groups.add(self)

        self._group_state_getter = state
        self._group_state_setter = state_set

        if commands:
            try:
                self.relay_commands = set(commands)
            except TypeError:
                self.relay_commands = _BagOfHolding()
        else:
            self.relay_commands = _Sieve()

        if command_send:
            if command_send is True:
                def dispatch_commands(items, command, *args, **kwargs):
                    for item in items:
                        item.command(command, *args, **kwargs)
                self.send_commands = _ImposterDict(dispatch_commands)
            elif command_send is False:
                self.send_commands = _Sieve()
            elif callable(command_send):
                self.send_commands = _ImposterDict(command_send)
            elif command_send is not None:
                self.send_commands = dict(command_send)

    def change_state(self, state):
        # TODO add a way to override this
        for item in self.members:
            item.change_state(state)

    @property
    def state(self):
        if self._group_state_getter:
            return self._group_state_getter(self.members)
        else:
            return super().state

    @state.setter
    def state(self, state):
        if self._group_state_setter:
            self._group_state_setter(self.members, state)
        else:
            # FIXME not sure if this will work
            super().state = state

    def command(self, command=None, *args, **kwargs):
        # Will receive any command by name
        if command and command in self.send_commands:
            self.send_commands[command](self.members, command, *args, **kwargs)

    def flattened(self, include_subgroups=False):
        """Return this group's members and all members of its subgroups, as a
        single list.

        Keyword arguments:
        include_subgroups -- If True, include each subgroup along with its
                             members. Otherwise, subgroups will not be
                             included but their members will be.
        """
        for item in self.members:
            if isinstance(item, Group):
                if include_subgroups:
                    yield item
                yield from item.flattened(include_subgroups)
            else:
                yield item

    def add(self, item):
        if item not in self.members:
            self.members.append(item)

    def _member_state_changed(self, member, state, source):
        if self._group_state_getter:
            post_event = event.StateChangeEvent(self, None, self.state, "group_member," + source, kind="after")
            self.idiotic.dispatcher.dispatch(post_event)

    def json(self):
        res = super().json()

        res.update({"members": [item.name for item in self.members]})
        return res



from idiotic import event, history
import idiotic
import logging

LOG = logging.getLogger("idiotic.scene")

class Scene:
    def __init__(self, name, active={}, inactive={}, tags=()):
        self.name = name
        self.history = history.History()
        self._active_state = dict(active)
        self._inactive_state = dict(inactive)
        self.tags = set(tags)
        self.tags.add("_scene")

        self.__active = False

        self._on_enter_funcs = []
        self._on_exit_funcs = []

        idiotic._register_scene(name, self)

    def _switch(self, val):
        if self.__active == val:
            LOG.debug("Ignoring redundant scene activation for {}".format(self))
            return

        pre_event = event.SceneEvent(self, val, "before")
        idiotic.dispatcher.dispatch(pre_event)
        if not pre_event.canceled:
            self.__active = val

            if hasattr(self, "history"):
                self.history.record(self.__active)

            if val:
                self._activate()
                for f in self._on_enter_funcs:
                    try:
                        f()
                    except:
                        pass
            else:
                self._deactivate()
                for f in self._on_exit_funcs:
                    try:
                        f()
                    except:
                        pass

            post_event = event.SceneEvent(self, val, "after")
            idiotic.dispatcher.dispatch(post_event)
        return val

    def _activate(self):
        itag = "_scene_" + self.name + "_inactive"
        atag = "_scene_" + self.name + "_active"
        for name in self._inactive_state.keys():
            try:
                item = idiotic.items[name]
                item.remove_state_overlay(tag=itag)
            except NameError:
                pass

        for name, state in self._active_state.items():
            try:
                item = idiotic.items[name]
                if isinstance(state, tuple):
                    state, disable = state
                else:
                    disable = False
                item.overlay_state(state, tag=atag, disable=disable)
            except NameError:
                pass

    def _deactivate(self):
        itag = "_scene_" + self.name + "_inactive"
        atag = "_scene_" + self.name + "_active"
        for name in self._active_state.keys():
            try:
                item = idiotic.items[name]
                item.remove_state_overlay(tag=atag)
            except NameError:
                pass

        for name, state in self._inactive_state.items():
            try:
                item = idiotic.items[name]
                if isinstance(state, tuple):
                    state, disable = state
                else:
                    disable = False
                item.overlay_state(state, tag=itag, disable=disable)
            except NameError:
                pass

    def enter(self):
        return self._switch(True)

    def exit(self):
        return self._switch(False)

    def on_exit(self, func):
        self._on_exit_funcs.append(func)
        return func

    def on_enter(self, func):
        self._on_enter_funcs.append(func)
        return func

    @property
    def active(self):
        return self.__active

    @active.setter
    def active(self, val):
        return self._switch(bool(val))

    def pack(self):
        result = {
            "__class__": type(self).__name__,
            "__owner__": getattr(self, 'MODULE', 'unknown'),
            "__kind__": "scene",
            "__active": self.__active,
            "__methods__": [k for k, v in self.__dict__.items() if callable(v)
                            and not k.startswith('__')]
        }
        result.update(self.__dict__)
        return result

    def json(self):
        return {
            "type": "Scene",
            "name": self.name,
            "active": self.__active,
            "tags": list(self.tags),
        }

    def __bool__(self):
        return self.__active

    def __str__(self):
        return "Scene {}".format(self.name)



from .api import _APIWrapper, join_url, jsonified, single_args
from .etc import mangle_name, IdioticEncoder
import functools
import logging
import imp
import sys
import os

LOG = logging.getLogger("idiotic.utils")

class AttrDict:
    def __init__(self, values={}):
        self.__values = dict(values)

    def _set(self, key, value):
        self.__values[mangle_name(key)] = value

    def all(self, filt=None):
        if isinstance(filt, Filter):
            return (v for v in self.__values.values() if filt.check(v))
        elif callable(filt):
            return (v for v in self.__values.values() if filt(v))
        else:
            return self.__values.values()

    def __getattr__(self, key):
        mangled = mangle_name(key)
        if mangled in self.__values:
            return self.__values[mangled]
        else:
            raise NameError("Could not find {} in AttrDict".format(key))

    def __getitem__(self, index):
        return getattr(self, index)

    def __contains__(self, key):
        return key in self.__values

class TaggedDict(AttrDict):
    def with_tags(self, tags):
        ts=set(tags)
        return self.all(filt=lambda i:ts.issubset(i.tags))

class BaseFilter:
    def __init__(self, *args):
        raise NotImplementedError()

    def __and__(self, other):
        return AndFilter(self, other)

    def __or__(self, other):
        return OrFilter(self, other)

    def __neg__(self):
        return NotFilter(self)

    def __invert__(self):
        return NotFilter(self)

    def __xor__(self, other):
        return XorFilter(self, other)

    def check(self, value):
        return True

class NotFilter(BaseFilter):
    def __init__(self, base):
        if not hasattr(base, "match"):
            raise ValueError("{} has no attribute 'match'".format(base))
        self.__base = base

    def check(self, event):
        return not self.__base.check(event)

    def __str__(self):
        return "NotFilter({})".format(str(self.__base))

    def __repr__(self):
        return "NotFilter({})".format(repr(self.__base))

class AndFilter(BaseFilter):
    def __init__(self, *bases):
        for f in bases:
            if not hasattr(f, "match"):
                raise ValueError("{} has no attribute 'match'".format(f))
        self.__bases = bases

    def match(self, event):
        return all((b.match(event) for b in self.__bases))

    def __str__(self):
        return "AndFilter({})".format(", ".join((str(b) for b in self.__bases)))

    def __repr__(self):
        return "AndFilter({})".format(", ".join((repr(b) for b in self.__bases)))

class OrFilter(BaseFilter):
    def __init__(self, *bases):
        for f in bases:
            if not hasattr(f, "match"):
                raise ValueError("{} has no attribute 'match'".format(f))
        self.__bases = bases

    def match(self, event):
        return any((b.match(event) for b in self.__bases))

    def __str__(self):
        return "OrFilter({})".format(", ".join((str(b) for b in self.__bases)))

    def __repr__(self):
        return "OrFilter({})".format(", ".join((repr(b) for b in self.__bases)))

class XorFilter(BaseFilter):
    def __init__(self, a, b):
        for f in (a, b):
            if not hasattr(f, "match"):
                raise ValueError("{} has no attribute 'match'".format(f))
        self.__a = a
        self.__b = b

    def match(self, event):
        return a.match(event) ^ b.match(event)

    def __str__(self):
        return "XorFilter({}, {})".format(str(a), str(b))

    def __repr__(self):
        return "XorFilter({}, {})".format(repr(a), repr(b))

class Filter(BaseFilter):
    def __init__(self, mode=None, filters=None, **kwargs):
        self.checks = []
        if mode is None:
            self.mode = all
        else:
            self.mode = mode

        if filters:
            # filters is used in case we need to use a reserved word
            # as an argument... though that should probably be avioded
            kwargs.update(filters)

        self.checks_def = kwargs

        for k, v in kwargs.items():
            # I'm very surprised I had to use a closure here. The only
            # other time I had to, I was doing some serious black
            # magic...
            def closure(k,v):
                if "__" in k:
                    key, op = k.rsplit("__", 1)
                else:
                    key, op = "", k
                path = key.split("__")

                if op == "contains":
                    self.checks.append(lambda e:v in self.__resolve_path(e, path))
                elif op == "not_contains":
                    self.checks.append(lambda e:v not in self.__resolve_path(e, path))
                elif op == "in":
                    self.checks.append(lambda e:self.__resolve_path(e, path) in v)
                elif op == "not_in":
                    self.checks.append(lambda e:self.__resolve_path(e, path) not in v)
                elif op == "is":
                    self.checks.append(lambda e:self.__resolve_path(e, path) is v)
                elif op == "is_not":
                    self.checks.append(lambda e:self.__resolve_path(e, path) is not v)
                elif op == "lt":
                    self.checks.append(lambda e:self.__resolve_path(e, path) < v)
                elif op == "gt":
                    self.checks.append(lambda e:self.__resolve_path(e, path) > v)
                elif op == "le":
                    self.checks.append(lambda e:self.__resolve_path(e, path) <= v)
                elif op == "ge":
                    self.checks.append(lambda e:self.__resolve_path(e, path) >= v)
                elif op == "ne":
                    self.checks.append(lambda e:self.__resolve_path(e, path) != v)
                elif op == "match":
                    self.checks.append(lambda e:v(self.__resolve_path(e, path)))
                elif op == "not_match":
                    self.checks.append(lambda e:not v(self.__resolve_path(e, path)))
                elif op == "eq":
                    self.checks.append(lambda e:self.__resolve_path(e, path) == v)
                elif op == "type":
                    self.checks.append(lambda e:type(self.__resolve_path(e, path)) == v)
                elif op == "type_not":
                    self.checks.append(lambda e:type(self.__resolve_path(e, path)) != v)
                elif op == "isinstance":
                    self.checks.append(lambda e:isinstance(self.__resolve_path(e, path), v))
                elif op == "not_isinstance":
                    self.checks.append(lambda e:not isinstance(self.__resolve_path(e, path), v))
                elif op == "hasattr":
                    self.checks.append(lambda e:hasattr(self.__resolve_path(e, path), v))
                elif op == "not_hasattr":
                    self.checks.append(lambda e:not hasattr(self.__resolve_path(e, path), v))
                else:
                    # By default just check for equality
                    path.append(op)
                    self.checks.append(lambda e:self.__resolve_path(e, path) == v)
            closure(k,v)

    def check(self, event):
        res = self.mode(c(event) for c in self.checks)
        return res

    def __resolve_path(self, e, path):
        # TODO make this function
        cur = e
        for key in path:
            if key:
                try:
                    cur = getattr(cur, key)
                except AttributeError:
                    return None
        return cur

    def __str__(self):
        return "Filter({})".format(", ".join(self.checks_def))

    def __repr__(self):
        return "Filter({})".format(", ".join(
            ("{}=<{}>".format(k.replace("__","."),repr(v)) for k,v in self.checks_def.items())))

def load_dir(path, include_assets=False):
    sys.path.insert(0, os.path.abspath("."))
    modules = []
    for f in os.listdir(path):
        try:
            if f.startswith(".") or f.endswith("~") or f.endswith("#") or f.startswith("__"):
                continue

            LOG.info("Loading file {}...".format(os.path.join(path, f)))
            name = os.path.splitext(f)[0]

            try:
                modules.append((imp.load_source(name, os.path.join(path, f)), None))
            except IsADirectoryError:
                LOG.info("Attempting to load directory {} as a module...".format(
                    os.path.join(path, f)))

                try:
                    mod = imp.load_source(name, os.path.join(path, f, '__init__.py'))
                    assets = None
                    if os.path.exists(os.path.join(path, f, 'assets')) and \
                       os.path.isdir(os.path.join(path, f, 'assets')):
                        assets = os.path.abspath(os.path.join(path, f, 'assets'))

                    modules.append((mod, assets))
                except FileNotFoundError:
                    LOG.error("Unable to load module {}: {} does not exist".format(
                        name, os.path.join(path, f, '__init__.py')))
        except:
            LOG.exception("Exception encountered while loading {}".format(os.path.join(path, f)))

    return modules

__ALL__ = [AttrDict, TaggedDict, mangle_name, IdioticEncoder, load_dir, _APIWrapper, join_url, jsonified, single_args]



from idiotic.utils import Filter
from asyncio import coroutine, iscoroutine, Queue, QueueFull
import logging
import functools

LOG = logging.getLogger("idiotic.dispatch")

class Dispatcher:
    def __init__(self):
        self.bindings = []
        self.queue = Queue()

    def bind(self, action, filt=Filter()):
        # The default filter will always return True
        self.bindings.append( (action, filt) )

    def unbind(self, action):
        for action, filt in list(self.bindings):
            self.bindings.remove( (action, filt) )
            return True
        return False

    def dispatch(self, event):
        for action in (a for a, f in self.bindings if f.check(event)):
            LOG.debug("Dispatching {}".format(str(action)))
            try:
                self.queue.put_nowait(functools.partial(action, event))
            except QueueFull:
                LOG.error("The unbounded queue is full! Pretty weird, eh?")

    @coroutine
    def run(self):
        while True:
            func = yield from self.queue.get()
            try:
                if not hasattr(func, "__name__"):
                    setattr(func, "__name__", "<unknown>")
                res = yield from coroutine(func)()

                while iscoroutine(res):
                    res = yield from res
            except:
                LOG.exception("Error while running {} from dispatch queue:".format(func))
